"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StatefulVerkleStateManager = void 0;
const common_1 = require("@ethereumjs/common");
const rlp_1 = require("@ethereumjs/rlp");
const util_1 = require("@ethereumjs/util");
const verkle_1 = require("@ethereumjs/verkle");
const debug_1 = require("debug");
const keccak_js_1 = require("ethereum-cryptography/keccak.js");
const originalStorageCache_js_1 = require("./cache/originalStorageCache.js");
const util_js_1 = require("./util.js");
class StatefulVerkleStateManager {
    constructor(opts) {
        /**
         * StateManager is run in DEBUG mode (default: false)
         * Taken from DEBUG environment variable
         *
         * Safeguards on debug() calls are added for
         * performance reasons to avoid string literal evaluation
         * @hidden
         */
        this.DEBUG = false;
        /**
         * Gets the account associated with `address` or `undefined` if account does not exist
         * @param address - Address of the `account` to get
         */
        this.getAccount = async (address) => {
            const elem = this._caches?.account?.get(address);
            if (elem !== undefined) {
                return elem.accountRLP !== undefined
                    ? (0, util_1.createPartialAccountFromRLP)(elem.accountRLP)
                    : undefined;
            }
            const stem = (0, util_1.getVerkleStem)(this.verkleCrypto, address, 0);
            // First retrieve the account "header" values from the trie
            const accountValues = await this._trie.get(stem, [
                util_1.VerkleLeafType.BasicData,
                util_1.VerkleLeafType.CodeHash,
            ]);
            let account;
            if (accountValues[0] !== undefined) {
                const basicData = (0, util_1.decodeVerkleLeafBasicData)(accountValues[0]);
                account = (0, util_1.createPartialAccount)({
                    version: basicData.version,
                    balance: basicData.balance,
                    nonce: basicData.nonce,
                    // Codehash is either untouched (i.e. undefined) or deleted (i.e. overwritten with zeros)
                    codeHash: accountValues[1] === undefined || (0, util_1.equalsBytes)(accountValues[1], new Uint8Array(32))
                        ? util_1.KECCAK256_NULL
                        : accountValues[1],
                    codeSize: basicData.codeSize,
                    storageRoot: util_1.KECCAK256_NULL, // TODO: Add storage stuff
                });
            }
            else if (accountValues[1] === undefined) {
                // account does not exist if both basic fields and codehash are undefined
                if (this.DEBUG) {
                    this._debug(`getAccount address=${address.toString()} from DB (non-existent)`);
                }
                this._caches?.account?.put(address, account);
            }
            if (this.DEBUG) {
                this._debug(`getAccount address=${address.toString()} stem=${(0, util_1.short)(stem)}`);
            }
            return account;
        };
        /**
         * Saves an account into state under the provided `address`.
         * @param address - Address under which to store `account`
         * @param account - The account to store or undefined if to be deleted
         */
        this.putAccount = async (address, account) => {
            if (this.DEBUG) {
                this._debug(`Save account address=${address} nonce=${account?.nonce} balance=${account?.balance} contract=${account && account.isContract() ? 'yes' : 'no'} empty=${account && account.isEmpty() ? 'yes' : 'no'}`);
            }
            if (this._caches?.account === undefined) {
                if (account !== undefined) {
                    const stem = (0, util_1.getVerkleStem)(this.verkleCrypto, address, 0);
                    const basicDataBytes = (0, util_1.encodeVerkleLeafBasicData)(account);
                    await this._trie.put(stem, [util_1.VerkleLeafType.BasicData, util_1.VerkleLeafType.CodeHash], [basicDataBytes, account.codeHash]);
                }
                else {
                    // Delete account
                    await this.deleteAccount(address);
                }
            }
            else {
                if (account !== undefined) {
                    this._caches?.account?.put(address, account, true);
                }
                else {
                    this._caches?.account?.del(address);
                }
            }
        };
        /**
         * Deletes an account from state under the provided `address`.
         * @param address - Address of the account which should be deleted
         */
        this.deleteAccount = async (address) => {
            if (this.DEBUG) {
                this._debug(`Delete account ${address}`);
            }
            this._caches?.deleteAccount(address);
            if (this._caches?.account === undefined) {
                const stem = (0, util_1.getVerkleStem)(this.verkleCrypto, address);
                // TODO: Determine the best way to clear code/storage for an account when deleting
                // Will need to inspect all possible code and storage keys to see if it's anything
                // other than untouched leaf values
                await this._trie.del(stem, [util_1.VerkleLeafType.BasicData, util_1.VerkleLeafType.CodeHash]);
            }
        };
        this.modifyAccountFields = async (address, accountFields) => {
            await (0, util_js_1.modifyAccountFields)(this, address, accountFields);
        };
        this.putCode = async (address, value) => {
            if (this.DEBUG) {
                this._debug(`putCode address=${address.toString()} value=${(0, util_1.short)(value)}`);
            }
            this._caches?.code?.put(address, value);
            const codeHash = (0, keccak_js_1.keccak256)(value);
            if ((0, util_1.equalsBytes)(codeHash, util_1.KECCAK256_NULL)) {
                // If the code hash is the null hash, no code has to be stored
                return;
            }
            if ((await this.getAccount(address)) === undefined) {
                await this.putAccount(address, new util_1.Account());
            }
            if (this.DEBUG) {
                this._debug(`Update codeHash (-> ${(0, util_1.short)(codeHash)}) for account ${address}`);
            }
            const codeChunks = (0, util_1.chunkifyCode)(value);
            const chunkStems = await (0, util_1.generateCodeStems)(codeChunks.length, address, this.verkleCrypto);
            const chunkSuffixes = (0, util_1.generateChunkSuffixes)(codeChunks.length);
            // Put the code chunks corresponding to the first stem (up to 128 chunks)
            await this._trie.put(chunkStems[0], chunkSuffixes.slice(0, codeChunks.length <= util_1.VERKLE_CODE_OFFSET ? codeChunks.length : util_1.VERKLE_CODE_OFFSET), codeChunks.slice(0, codeChunks.length <= util_1.VERKLE_CODE_OFFSET ? codeChunks.length : util_1.VERKLE_CODE_OFFSET));
            // Put additional chunks under additional stems as applicable
            for (let stem = 1; stem < chunkStems.length; stem++) {
                const sliceStart = util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * (stem - 1);
                const sliceEnd = value.length <= util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * stem
                    ? value.length
                    : util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * stem;
                await this._trie.put(chunkStems[stem], chunkSuffixes.slice(sliceStart, sliceEnd), codeChunks.slice(sliceStart, sliceEnd));
            }
            await this.modifyAccountFields(address, { codeHash, codeSize: value.length });
        };
        this.getCode = async (address) => {
            if (this.DEBUG) {
                this._debug(`getCode address=${address.toString()}`);
            }
            const elem = this._caches?.code?.get(address);
            if (elem !== undefined) {
                return elem.code ?? new Uint8Array(0);
            }
            const account = await this.getAccount(address);
            if (!account) {
                return new Uint8Array(0);
            }
            if (!account.isContract()) {
                return new Uint8Array(0);
            }
            // allocate the code
            const codeSize = account.codeSize;
            const stems = await (0, util_1.generateCodeStems)(Math.ceil(codeSize / util_1.VERKLE_CODE_CHUNK_SIZE), address, this.verkleCrypto);
            const chunkSuffixes = (0, util_1.generateChunkSuffixes)(Math.ceil(codeSize / util_1.VERKLE_CODE_CHUNK_SIZE));
            const chunksByStem = new Array(stems.length);
            // Retrieve the code chunks stored in the first leaf node
            chunksByStem[0] = await this._trie.get(stems[0], chunkSuffixes.slice(0, codeSize <= util_1.VERKLE_CODE_OFFSET ? codeSize : util_1.VERKLE_CODE_OFFSET));
            // Retrieve code chunks on any additional stems
            for (let stem = 1; stem < stems.length; stem++) {
                const sliceStart = util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * (stem - 1);
                const sliceEnd = codeSize <= util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * stem
                    ? codeSize
                    : util_1.VERKLE_CODE_OFFSET + util_1.VERKLE_NODE_WIDTH * stem;
                chunksByStem[stem] = await this._trie.get(stems[stem], chunkSuffixes.slice(sliceStart, sliceEnd));
            }
            const chunks = chunksByStem.flat();
            const code = new Uint8Array(codeSize);
            // Insert code chunks into final array (skipping PUSHDATA overflow indicator byte)
            for (let x = 0; x < chunks.length; x++) {
                if (chunks[x] === undefined)
                    throw new Error(`expected code chunk at ID ${x}, got undefined`);
                // Determine code ending byte (if we're on the last chunk)
                let sliceEnd = 32;
                if (x === chunks.length - 1) {
                    sliceEnd = (codeSize % util_1.VERKLE_CODE_CHUNK_SIZE) + 1;
                }
                code.set(chunks[x].slice(1, sliceEnd), code.byteOffset + x * util_1.VERKLE_CODE_CHUNK_SIZE);
            }
            this._caches?.code?.put(address, code);
            return code;
        };
        this.getCodeSize = async (address) => {
            const accountBytes = (await this._trie.get((0, util_1.getVerkleStem)(this.verkleCrypto, address), [util_1.VerkleLeafType.BasicData]))[0];
            if (accountBytes === undefined)
                return 0;
            return (0, util_1.decodeVerkleLeafBasicData)(accountBytes).codeSize;
        };
        this.getStorage = async (address, key) => {
            if (key.length !== 32) {
                throw new Error('Storage key must be 32 bytes long');
            }
            const cachedValue = this._caches?.storage?.get(address, key);
            if (cachedValue !== undefined) {
                const decoded = rlp_1.RLP.decode(cachedValue ?? new Uint8Array(0));
                return decoded;
            }
            const account = await this.getAccount(address);
            if (!account) {
                return new Uint8Array();
            }
            const storageKey = await (0, util_1.getVerkleTreeKeyForStorageSlot)(address, (0, util_1.bytesToBigInt)(key, true), this.verkleCrypto);
            const value = await this._trie.get(storageKey.slice(0, 31), [storageKey[31]]);
            this._caches?.storage?.put(address, key, value[0] ?? (0, util_1.hexToBytes)('0x80'));
            const decoded = rlp_1.RLP.decode(value[0] ?? new Uint8Array(0));
            return decoded;
        };
        this.putStorage = async (address, key, value) => {
            value = (0, util_1.unpadBytes)(value);
            this._caches?.storage?.put(address, key, rlp_1.RLP.encode(value));
            if (this._caches?.storage === undefined) {
                const storageKey = await (0, util_1.getVerkleTreeKeyForStorageSlot)(address, (0, util_1.bytesToBigInt)(key, true), this.verkleCrypto);
                await this._trie.put(storageKey.slice(0, 31), [storageKey[31]], [rlp_1.RLP.encode(value)]);
            }
        };
        this.clearStorage = async (address) => {
            // TODO: Determine if it's possible to clear the actual slots in the trie
            // since the EIP doesn't seem to state how to handle this
            // The main concern I have is that we have no way of identifying all storage slots
            // for a given account so we can't correctly update the trie's root hash
            // (since presumably "clearStorage" would imply writing over all of the storage slots with zeros)
            // Also, do we still need a storageRoot? - presumably not since we don't have separate storage tries
            this._caches?.storage?.clearStorage(address);
        };
        this.checkpoint = async () => {
            this._trie.checkpoint();
            this._caches?.checkpoint();
            this._checkpointCount++;
        };
        this.commit = async () => {
            await this._trie.commit();
            this._caches?.commit();
            this._checkpointCount--;
            if (this._checkpointCount === 0) {
                await this.flush();
                this.originalStorageCache.clear();
            }
            if (this.DEBUG) {
                this._debug(`state checkpoint committed`);
            }
        };
        this.revert = async () => {
            await this._trie.revert();
            this._caches?.revert();
            this._checkpointCount--;
            if (this._checkpointCount === 0) {
                await this.flush();
                this.originalStorageCache.clear();
            }
        };
        this.flush = async () => {
            const codeItems = this._caches?.code?.flush() ?? [];
            for (const item of codeItems) {
                const addr = (0, util_1.createAddressFromString)(`0x${item[0]}`);
                const code = item[1].code;
                if (code === undefined) {
                    continue;
                }
                await this.putCode(addr, code);
            }
            const storageItems = this._caches?.storage?.flush() ?? [];
            for (const item of storageItems) {
                const address = (0, util_1.createAddressFromString)(`0x${item[0]}`);
                const keyHex = item[1];
                const keyBytes = (0, util_1.unprefixedHexToBytes)(keyHex);
                const value = item[2];
                const decoded = rlp_1.RLP.decode(value ?? new Uint8Array(0));
                const account = await this.getAccount(address);
                if (account) {
                    await this.putStorage(address, keyBytes, decoded);
                }
            }
            const accountItems = this._caches?.account?.flush() ?? [];
            for (const item of accountItems) {
                const address = (0, util_1.createAddressFromString)(`0x${item[0]}`);
                const elem = item[1];
                if (elem.accountRLP === undefined) {
                    await this.deleteAccount(address);
                }
                else {
                    const account = (0, util_1.createPartialAccountFromRLP)(elem.accountRLP);
                    await this.putAccount(address, account);
                }
            }
        };
        // Skip DEBUG calls unless 'ethjs' included in environmental DEBUG variables
        // Additional window check is to prevent vite browser bundling (and potentially other) to break
        this.DEBUG =
            typeof window === 'undefined' ? (process?.env?.DEBUG?.includes('ethjs') ?? false) : false;
        this._checkpointCount = 0;
        if (opts.common?.isActivatedEIP(6800) === false)
            throw new Error('EIP-6800 required for verkle state management');
        this.common = opts.common ?? new common_1.Common({ chain: common_1.Mainnet, eips: [6800] });
        this._trie =
            opts.trie ??
                new verkle_1.VerkleTree({ verkleCrypto: opts.verkleCrypto, db: new util_1.MapDB() });
        this._debug = (0, debug_1.default)('statemanager:verkle:stateful');
        this.originalStorageCache = new originalStorageCache_js_1.OriginalStorageCache(this.getStorage.bind(this));
        this._caches = opts.caches;
        this.keccakFunction = opts.common?.customCrypto.keccak256 ?? keccak_js_1.keccak256;
        this.verkleCrypto = opts.verkleCrypto;
    }
    getStateRoot() {
        throw new Error('Method not implemented.');
    }
    setStateRoot(_stateRoot, _clearCache) {
        throw new Error('Method not implemented.');
    }
    hasStateRoot(_root) {
        throw new Error('Method not implemented.');
    }
    dumpStorage(_address) {
        throw new Error('Method not implemented.');
    }
    dumpStorageRange(_address, _startKey, _limit) {
        throw new Error('Method not implemented.');
    }
    clearCaches() {
        throw new Error('Method not implemented.');
    }
    shallowCopy(_downlevelCaches) {
        throw new Error('Method not implemented.');
    }
    async checkChunkWitnessPresent(_address, _codeOffset) {
        throw new Error('Method not implemented.');
    }
}
exports.StatefulVerkleStateManager = StatefulVerkleStateManager;
//# sourceMappingURL=statefulVerkleStateManager.js.map